#include "Brute.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "FastPoints.hpp"

using std::vector;
using std::pair;

Brute::Brute(const Problem& problem,
             const FastPoints& fastPoints,
             const vector<Point>& points,
             OnSolution onSolution,
             OnDebug onDebug) :
    m_problem(problem),
    m_fastPoints(fastPoints),
    m_points(points),
    m_onSolution(onSolution),
    m_onDebug(onDebug)
{
    m_solution.m_vertices.resize(problem.m_vertices.size());

    m_vertexPoints.resize(problem.m_vertices.size());
    m_backupPoints.resize(problem.m_edges.size());
}

Brute::~Brute()
{
}

bool Brute::solve(uint64_t maxNodes,
                  size_t firstVertex,
                  const vector<Point>* pTargets)
{
    m_pTargets = pTargets;
    m_maxNodes = maxNodes;
    m_nodeCount = 0;
    for (auto& vertex : m_solution.m_vertices)
    {
        vertex = Point(-999, -999);
    }
    m_processed.clear();
    m_processed.resize(m_problem.m_vertices.size());

    for (auto& vertexPoints : m_vertexPoints)
    {
        vertexPoints.clear();
    }

    m_firstVertex = firstVertex;

    return recurse(m_solution.m_vertices.size(), (size_t)-1);
}

bool Brute::recurse(size_t depth, size_t lastVertex)
{
    //printf("> %" PRIuZ "\n", depth);
    m_nodeCount++;
    if ((m_nodeCount & 0xfff) == 0)
    {
        if (m_onDebug)
        {
            if (m_onDebug(m_solution))
            {
                return true;
            }
        }
    }

    if (m_maxNodes != 0 && m_nodeCount >= m_maxNodes)
    {
        return true;
    }

    if (depth == 0)
    {
#if 1
        if (!isValid(m_problem, m_solution))
        {
            printf("Not valid!\n");
            exit(1);
        }
#endif
        uint64_t score = calcScore(m_problem, m_solution);
        //printf("Found node count = %" PRIu64 "\n", m_nodeCount);
        //printf("< %" PRIuZ "\n", depth);
        return m_onSolution(m_solution, score);
    }

    size_t bestVertex = 0;
    size_t bestOkCount = (size_t)-1;
    if (lastVertex == (size_t)-1)
    {
        bestVertex = m_firstVertex;
        bestOkCount = m_points.size();
    }
    else
    {
        for (auto& neighbor : m_fastPoints.getNeighbors(lastVertex))
        {
            if (m_processed[neighbor.m_vertexIdx]) continue;
            if (m_fastPoints.getEdgeInfo(neighbor.m_edgeIdx).m_duplicate) continue;
            auto& vertexPoints = m_vertexPoints[neighbor.m_vertexIdx];
            auto& backupPoints = m_backupPoints[neighbor.m_edgeIdx];
            vertexPoints.swap(backupPoints);
            if (bestOkCount == 0) continue;
            vertexPoints.clear();
            if (backupPoints.empty())
            {
                m_fastPoints.getPoints(neighbor.m_edgeIdx,
                                       m_solution.m_vertices[lastVertex],
                                       &m_pointsTmp);
                for (auto point : m_pointsTmp)
                {
                    if (!edgeInPolygon(m_problem.m_hole,
                                       m_solution.m_vertices[lastVertex],
                                       point))
                    {
                        continue;
                    }
                    vertexPoints.push_back(point);
                }

#if 0
                if (m_pTargets)
                {
                    Point target = (*m_pTargets)[neighbor.m_vertexIdx];
                    auto targetComp = [&](Point a, Point b) -> bool
                    {
                        uint64_t d2A = calcD(a, target);
                        uint64_t d2B = calcD(b, target);
                        if (d2A != d2B) return d2A < d2B;
                        return a < b;
                    };
                    std::sort(vertexPoints.begin(),
                              vertexPoints.end(),
                              targetComp);
                }
                else
                {
                    std::sort(vertexPoints.begin(),
                              vertexPoints.end());
                }
#endif
            }
            else
            {
                uint64_t epsilon = m_problem.m_epsilon;
                for (auto point : backupPoints)
                {
                    uint64_t dOld = m_fastPoints.getEdgeInfo(neighbor.m_edgeIdx).m_d2;
                    uint64_t dNew = calcD(m_solution.m_vertices[lastVertex],
                                          point);
                    if (!isValidD(dOld, dNew, epsilon))
                    {
                        continue;
                    }
                    if (!edgeInPolygon(m_problem.m_hole,
                                       m_solution.m_vertices[lastVertex],
                                       point))
                    {
                        continue;
                    }
                    vertexPoints.push_back(point);
                }
            }

            if (vertexPoints.empty())
            {
                bestVertex = neighbor.m_vertexIdx;
                bestOkCount = 0;
            }
        }

#if 0
        for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
        {
            if (m_processed[iVertex]) continue;
            printf("%2" PRIuZ ": %" PRIuZ "\n", iVertex, m_vertexPoints[iVertex].size());
        }
#endif

        if (bestOkCount != 0)
        {
            for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
            {
                if (m_processed[iVertex]) continue;
                auto& vertexPoints = m_vertexPoints[iVertex];
                if (vertexPoints.empty()) continue;
                size_t okCount = vertexPoints.size();

                if (okCount < bestOkCount)
                {
                    bestVertex = iVertex;
                    bestOkCount = okCount;
                }
            }
        }
    }

    //printf("bestOkCount = %" PRIuZ "\n", bestOkCount);
    //printf("bestVertex = %" PRIuZ "\n", bestVertex);
    //if (bestOkCount == 0) return false;

    if (bestOkCount != 0)
    {
        auto& points =
            lastVertex == (size_t)-1 ?
            m_points :
            m_vertexPoints[bestVertex];
        size_t numPoints = points.size();

#if 1
        vector<pair<uint64_t, Point>> greedyPoints(numPoints);
        if (m_pTargets)
        {
            Point target = (*m_pTargets)[bestVertex];
            if (target.x == -999)
            {
                size_t numCorners = m_problem.m_hole.size();
                vector<uint64_t> corners(numCorners);
                for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
                {
                    uint64_t minD = std::numeric_limits<uint64_t>::max();
                    for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
                    {
                        if (!m_processed[iVertex]) continue;
                        minD = std::min(minD, calcD(m_problem.m_hole[iCorner],
                                                    m_solution.m_vertices[iVertex]));
                    }
                    corners[iCorner] = minD;
                }

                for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
                {
                    Point point = points[iPoint];
                    uint64_t score = 0;
                    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
                    {
                        uint64_t minD = corners[iCorner];
                        minD = std::min(minD, calcD(m_problem.m_hole[iCorner],
                                                    point));
                        score += minD;
                    }
                    greedyPoints[iPoint].first = score;
                    greedyPoints[iPoint].second = point;
                }
            }
            else
            {
                for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
                {
                    Point point = points[iPoint];
                    uint64_t score = calcD(target, point);
                    greedyPoints[iPoint].first = score;
                    greedyPoints[iPoint].second = point;
                }
            }
        }
        else
        {
            for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
            {
                Point point = points[iPoint];
                uint64_t score = 0;
                greedyPoints[iPoint].first = score;
                greedyPoints[iPoint].second = point;
            }
        }
        std::sort(greedyPoints.begin(), greedyPoints.end());
#endif

        m_processed[bestVertex] = 1;
        for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
        {
#if 0
            if (iPoint == 3) break;
#endif
            Point point = greedyPoints[iPoint].second;
            m_solution.m_vertices[bestVertex] = point;
            if (recurse(depth - 1, bestVertex))
            {
                return true;
            }
            m_solution.m_vertices[bestVertex] = Point(-999, -999);
        }
        m_processed[bestVertex] = 0;
    }

    if (lastVertex != (size_t)-1)
    {
        for (auto& neighbor : m_fastPoints.getNeighbors(lastVertex))
        {
            if (m_processed[neighbor.m_vertexIdx]) continue;
            if (m_fastPoints.getEdgeInfo(neighbor.m_edgeIdx).m_duplicate) continue;
            auto& vertexPoints = m_vertexPoints[neighbor.m_vertexIdx];
            auto& backupPoints = m_backupPoints[neighbor.m_edgeIdx];
            vertexPoints.swap(backupPoints);
        }
    }

    //printf("< %" PRIuZ "\n", depth);
    return false;
}
