#include "DrawProblem.hpp"
#include "Graphics.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"
#include "FastPoints.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

void drawProblem(Graphics& gr,
                 const Problem& problem,
                 const FastPoints& fastPoints,
                 const std::vector<Point>& points,
                 const Solution& solution,
                 const string& text)
{
    int64_t minX = std::numeric_limits<int64_t>::max();
    int64_t maxX = std::numeric_limits<int64_t>::min();
    int64_t minY = std::numeric_limits<int64_t>::max();
    int64_t maxY = std::numeric_limits<int64_t>::min();
    for (auto& corner : problem.m_hole)
    {
        minX = std::min(minX, corner.x);
        maxX = std::max(maxX, corner.x);
        minY = std::min(minY, corner.y);
        maxY = std::max(maxY, corner.y);
    }

    uint32_t width = gr.getWidth();
    uint32_t height = gr.getHeight();

    gr.setFont(FontSpacing::Monospace, FontFace::SansSerif,
               18, false, false);
    uint32_t fontHeight = gr.getFontHeight();

    int32_t zoom = 1;
    while ((int32_t)(maxX - minX + 2) * (zoom + 1) + 1 <= (int32_t)width &&
           (int32_t)(maxY - minY + 2) * (zoom + 1) + 1 <= (int32_t)height)
    {
        zoom++;
    }

    gr.fillRect(0, 0, width, height, 0x404040);

    if (!text.empty())
    {
        gr.drawText(2, 0,
                    text.c_str(),
                    0xffffff,
                    HAlign::Left, VAlign::Top);
    }

    if (problem.m_vertices.empty())
    {
        return;
    }

    auto getX = [&](int64_t x) -> int32_t
    {
        return (int32_t)(x - minX + 1) * zoom;
    };

    auto getY = [&](int64_t y) -> int32_t
    {
        return (int32_t)(y - minY + 1) * zoom + (int32_t)fontHeight;
    };

    size_t numCorners = problem.m_hole.size();

    vector<GPoint> gpoints(numCorners);
    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
    {
        gpoints[iCorner].x = getX(problem.m_hole[iCorner].x);
        gpoints[iCorner].y = getY(problem.m_hole[iCorner].y);
    }

    gr.fillPolygon(gpoints.data(), numCorners, 0x606060);
    gr.drawPolygon(gpoints.data(), numCorners, 0xffffff);

    if (zoom > 1)
    {
        for (Point point : points)
        {
            if (zoom > 7)
            {
                gr.fillCircle(getX(point.x) - 1,
                              getY(point.y) - 1,
                              3, 3,
                              0xff0000);
            }
            else if (zoom > 5)
            {
                gr.fillCircle(getX(point.x),
                              getY(point.y),
                              2, 2,
                              0xff0000);
            }
            else
            {
                gr.drawPoint(getX(point.x), getY(point.y), 0xff0000);
            }
        }
    }

    {
        vector<Point> curPoints;
        fastPoints.getPoints(0,
                             problem.m_vertices[problem.m_edges[0].b],
                             &curPoints);

        for (Point point : curPoints)
        {
            gr.fillCircle(getX(point.x) - 1,
                          getY(point.y) - 1,
                          3, 3,
                          0xffff00);
        }
    }

    for (auto& edge : problem.m_edges)
    {
        Point pa = problem.m_vertices[edge.a];
        Point pb = problem.m_vertices[edge.b];
        uint32_t color =
            edgeInPolygon(problem.m_hole, pa, pb) ? 0xff0000 : 0xffff00;
        gr.drawLine(getX(pa.x), getY(pa.y),
                    getX(pb.x), getY(pb.y),
                    color);
    }

    if (!solution.m_vertices.empty())
    {
        for (auto& point : solution.m_vertices)
        {
            if (point != Point(-999, -999))
            {
                gr.fillCircle(getX(point.x) - 3,
                              getY(point.y) - 3,
                              7, 7,
                              0x00ffff);
            }
        }

        gr.setLineWidth(3);
        for (auto& edge : problem.m_edges)
        {
            Point pa = solution.m_vertices[edge.a];
            Point pb = solution.m_vertices[edge.b];
            if (pa != Point(-999, -999) &&
                pb != Point(-999, -999))
            {
                gr.drawLine(getX(pa.x), getY(pa.y),
                            getX(pb.x), getY(pb.y),
                            0x00ffff);
            }
        }
    }

#if 1
    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
    {
        gr.drawText(getX(problem.m_hole[iCorner].x),
                    getY(problem.m_hole[iCorner].y),
                    strprintf("%" PRIuZ "", iCorner).c_str(),
                    0xffffff,
                    HAlign::Left, VAlign::Top);
    }

    for (size_t iVertex = 0; iVertex < problem.m_vertices.size(); iVertex++)
    {
        gr.drawText(getX(problem.m_vertices[iVertex].x),
                    getY(problem.m_vertices[iVertex].y),
                    strprintf("%" PRIuZ "", iVertex).c_str(),
                    0xffff00,
                    HAlign::Left, VAlign::Top);
    }
#endif
}
