#include "FastPoints.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"

FastPoints::FastPoints()
{
}

FastPoints::~FastPoints()
{
}

void FastPoints::init(const Problem& problem,
                      const std::vector<Point>& points)
{
    m_minX = std::numeric_limits<int64_t>::max();
    m_maxX = std::numeric_limits<int64_t>::min();
    m_minY = std::numeric_limits<int64_t>::max();
    m_maxY = std::numeric_limits<int64_t>::min();
    for (Point p : problem.m_hole)
    {
        m_minX = std::min(m_minX, p.x);
        m_maxX = std::max(m_maxX, p.x);
        m_minY = std::min(m_minY, p.y);
        m_maxY = std::max(m_maxY, p.y);
    }
    m_stride = m_maxX - m_minX + 1;
    m_grid.clear();
    m_grid.resize(m_stride * (m_maxY - m_minY + 1));

#if 0
    for (int64_t y = m_minY; y <= m_maxY; y++)
    {
        for (int64_t x = m_minX; x <= m_maxX; x++)
        {
            Point point(x, y);
            if (pointInPolygon(problem.m_hole, point))
            {
                m_grid[(y - m_minY) * m_stride + (x - m_minX)] |= 1;
            }
        }
    }
#endif

#if 1
    for (Point p : points)
    {
        if (p.x >= m_minX &&
            p.x <= m_maxX &&
            p.y >= m_minY &&
            p.y <= m_maxY)
        {
            m_grid[(p.y - m_minY) * m_stride + (p.x - m_minX)] |= 1;
        }
    }
#endif

    for (Point p : problem.m_hole)
    {
        if (p.x >= m_minX &&
            p.x <= m_maxX &&
            p.y >= m_minY &&
            p.y <= m_maxY)
        {
            m_grid[(p.y - m_minY) * m_stride + (p.x - m_minX)] |= 2;
        }
    }

    m_edgeInfo.resize(problem.m_edges.size());
    m_edgeMap.resize(problem.m_edges.size());
    m_neighbors.resize(problem.m_vertices.size());
    for (size_t edgeIdx = 0; edgeIdx < problem.m_edges.size(); edgeIdx++)
    {
        auto& edge = problem.m_edges[edgeIdx];
        {
            auto& neighbor = m_neighbors[edge.a].emplace_back();
            neighbor.m_edgeIdx = edgeIdx;
            neighbor.m_vertexIdx = edge.b;
        }
        {
            auto& neighbor = m_neighbors[edge.b].emplace_back();
            neighbor.m_edgeIdx = edgeIdx;
            neighbor.m_vertexIdx = edge.a;
        }
        uint64_t dist = calcD(problem.m_vertices[edge.a],
                              problem.m_vertices[edge.b]);
        auto& edgeInfo = m_edgeInfo[edgeIdx];
        edgeInfo.m_d2 = dist;
        {
            int64_t d1 = 0;
            while ((d1 + 1) * (d1 + 1) * 1000000 <= (1000000 + (int64_t)problem.m_epsilon) * (int64_t)dist)
            {
                d1++;
            }
            edgeInfo.m_maxParallelD1 = (uint64_t)d1;
        }
        {
            int64_t d1 = 0;
            while (d1 * d1 * 1000000 < (1000000 + (int64_t)problem.m_epsilon) * (int64_t)dist)
            {
                d1++;
            }
            d1 *= 1000;
            while (d1 > 0 && (d1 - 1) * (d1 - 1) >= (1000000 + (int64_t)problem.m_epsilon) * (int64_t)dist)
            {
                d1--;
            }
            edgeInfo.m_maxD1_1000 = (uint64_t)d1;
        }
        edgeInfo.m_duplicate = false;
        for (size_t jEdge = 0; jEdge < edgeIdx; jEdge++)
        {
            if ((problem.m_edges[edgeIdx].a == problem.m_edges[jEdge].a &&
                 problem.m_edges[edgeIdx].b == problem.m_edges[jEdge].b) ||
                (problem.m_edges[edgeIdx].a == problem.m_edges[jEdge].b &&
                 problem.m_edges[edgeIdx].b == problem.m_edges[jEdge].a))
            {
                edgeInfo.m_duplicate = true;
            }
        }
        size_t distIdx = 0;
        while (distIdx < m_distInfo.size() &&
               m_distInfo[distIdx].m_dist != dist)
        {
            distIdx++;
        }
        m_edgeMap[edgeIdx] = distIdx;
        if (distIdx < m_distInfo.size())
        {
            continue;
        }
        auto& distInfo = m_distInfo.emplace_back();
        distInfo.m_dist = dist;
        int64_t low = (1000000 - (int64_t)problem.m_epsilon) * (int64_t)dist;
        int64_t high = (1000000 + (int64_t)problem.m_epsilon) * (int64_t)dist;
        distInfo.m_zero = 0 >= low;
        int64_t limit = 0;
        {
            int64_t i = 1;
            while (i * i * 1000000 < low)
            {
                i++;
            }
            while (i * i * 1000000 <= high)
            {
                distInfo.m_parallel.push_back(i);
                i++;
            }
            limit = i;
        }
        {
            int64_t i = 1;
            while (2 * i * i * 1000000 < low)
            {
                i++;
            }
            while (2 * i * i * 1000000 <= high)
            {
                distInfo.m_diagonal.push_back(i);
                i++;
            }
        }
        for (int64_t i = 2; i < limit; i++)
        {
            for (int64_t j = 1; j < i; j++)
            {
                if ((i * i + j * j) * 1000000 >= low &&
                    (i * i + j * j) * 1000000 <= high)
                {
                    distInfo.m_between.push_back(Point(i, j));
                }
            }
        }
    }

    {
        size_t N = problem.m_vertices.size();
        m_numVertices = N;
        m_shortestPaths.resize(N * N);
        for (size_t i = 0; i < N; i++)
        {
            for (size_t j = 0; j < N; j++)
            {
                m_shortestPaths[i * N + j] = ((uint64_t)-1) >> 1;
            }
        }
        for (size_t i = 0; i < N; i++)
        {
            m_shortestPaths[i * N + i] = 0;
        }
        for (size_t edgeIdx = 0; edgeIdx < problem.m_edges.size(); edgeIdx++)
        {
            auto& edge = problem.m_edges[edgeIdx];
            size_t i = edge.a;
            size_t j = edge.b;
            auto& edgeInfo = m_edgeInfo[edgeIdx];
            m_shortestPaths[i * N + j] = edgeInfo.m_maxD1_1000;
            m_shortestPaths[j * N + i] = edgeInfo.m_maxD1_1000;
        }
        for (size_t k = 0; k < N; k++)
        {
            for (size_t i = 0; i < N; i++)
            {
                for (size_t j = 0; j < N; j++)
                {
                    m_shortestPaths[i * N + j] =
                        std::min(m_shortestPaths[i * N + j],
                                 m_shortestPaths[i * N + k] +
                                 m_shortestPaths[k * N + j]);
                }
            }
        }

#if 0
        for (size_t i = 0; i < N; i++)
        {
            for (size_t j = 0; j < N; j++)
            {
                printf("  %7" PRIu64 "", m_shortestPaths[i * N + j]);
            }
            printf("\n");
        }
        printf("\n");
#endif
    }
}

void FastPoints::getPoints(size_t edgeIdx,
                           Point point,
                           std::vector<Point>* pPoints) const
{
    size_t distIdx = m_edgeMap[edgeIdx];
    auto& distInfo = m_distInfo[distIdx];
    pPoints->clear();
    auto add = [&](Point p)
    {
        p += point;
        if (isInGrid(p))
        {
            pPoints->push_back(p);
        }
    };

    if (distInfo.m_zero)
    {
        add(Point(0, 0));
    }
    for (int64_t i : distInfo.m_parallel)
    {
        add(Point(i, 0));
        add(Point(-i, 0));
        add(Point(0, i));
        add(Point(0, -i));
    }
    for (int64_t i : distInfo.m_diagonal)
    {
        add(Point(i, i));
        add(Point(-i, i));
        add(Point(i, -i));
        add(Point(-i, -i));
    }
    for (Point p : distInfo.m_between)
    {
        add(Point(p.x, p.y));
        add(Point(-p.x, p.y));
        add(Point(p.x, -p.y));
        add(Point(-p.x, -p.y));
        add(Point(p.y, p.x));
        add(Point(-p.y, p.x));
        add(Point(p.y, -p.x));
        add(Point(-p.y, -p.x));
    }

    //std::sort(pPoints->begin(), pPoints->end());
}
