#include "Problem.hpp"
#include "Json.hpp"
#include "JsonUtils.hpp"
#include "FileUtils.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

namespace J = rapidjson;

namespace
{
    void parsePoint(const J::Value& jPoint,
                    Point* pPoint)
    {
        auto& point = *pPoint;

        JCheckArray(jPoint);
        if (jPoint.Size() != 2)
        {
            throw JException("Invalid point size");
        }
        point.x = JGetInt64(jPoint[0]);
        point.y = JGetInt64(jPoint[1]);
    }

    void parsePoints(const J::Value& jPoints,
                     vector<Point>* pPoints)
    {
        auto& points = *pPoints;

        JCheckArray(jPoints);

        points.clear();
        points.reserve(jPoints.Size());
        for (auto& jPoint : jPoints.GetArray())
        {
            auto& point = points.emplace_back();
            parsePoint(jPoint, &point);
        }
    }

    void parseEdge(const J::Value& jEdge,
                   Edge* pEdge)
    {
        auto& edge = *pEdge;

        JCheckArray(jEdge);
        if (jEdge.Size() != 2)
        {
            throw JException("Invalid edge size");
        }
        edge.a = JGetUint64(jEdge[0]);
        edge.b = JGetUint64(jEdge[1]);
    }

    void parseEdges(const J::Value& jEdges,
                    vector<Edge>* pEdges)
    {
        auto& edges = *pEdges;

        JCheckArray(jEdges);

        edges.clear();
        edges.reserve(jEdges.Size());
        for (auto& jEdge : jEdges.GetArray())
        {
            auto& edge = edges.emplace_back();
            parseEdge(jEdge, &edge);
        }
    }

    void parseBonusType(const J::Value& jBonusType,
                        BonusType* pBonusType)
    {
        auto& bonusType = *pBonusType;

        string strBonusType = JGetString(jBonusType);
        if (strBonusType == "GLOBALIST")
        {
            bonusType = BonusType::Globalist;
        }
        else if (strBonusType == "BREAK_A_LEG")
        {
            bonusType = BonusType::BreakALeg;
        }
        else if (strBonusType == "WALLHACK")
        {
            bonusType = BonusType::WallHack;
        }
        else if (strBonusType == "SUPERFLEX")
        {
            bonusType = BonusType::Superflex;
        }
        else
        {
            throw JException("Unknown bonus type");
        }
    }

    void parseBonus(const J::Value& jBonus,
                    Bonus* pBonus)
    {
        auto& bonus = *pBonus;

        JCheckObject(jBonus);

        {
            auto& jPosition = JObjectGetArray(jBonus, "position");
            parsePoint(jPosition, &bonus.m_position);
        }

        {
            auto& jBonusType = JObjectGet(jBonus, "bonus");
            parseBonusType(jBonusType, &bonus.m_bonusType);
        }

        bonus.m_id = JObjectGetUint64(jBonus, "problem");
    }

    void parseBonuses(const J::Value& jBonuses,
                      vector<Bonus>* pBonuses)
    {
        auto& bonuses = *pBonuses;

        JCheckArray(jBonuses);

        bonuses.clear();
        bonuses.reserve(jBonuses.Size());
        for (auto& jBonus : jBonuses.GetArray())
        {
            auto& bonus = bonuses.emplace_back();
            parseBonus(jBonus, &bonus);
        }
    }

    void parseProblem(const J::Value& jProblem,
                      Problem* pProblem)
    {
        auto& problem = *pProblem;

        JCheckObject(jProblem);

        {
            auto& jHole = JObjectGetArray(jProblem, "hole");
            parsePoints(jHole, &problem.m_hole);
        }

        auto& jFigure = JObjectGetObject(jProblem, "figure");

        {
            auto& jEdges = JObjectGetArray(jFigure, "edges");
            parseEdges(jEdges, &problem.m_edges);
        }

        {
            auto& jVertices = JObjectGetArray(jFigure, "vertices");
            parsePoints(jVertices, &problem.m_vertices);
        }

        problem.m_epsilon = JObjectGetUint64(jProblem, "epsilon");

        if (jProblem.HasMember("bonuses"))
        {
            auto& jBonuses = JObjectGetArray(jProblem, "bonuses");
            parseBonuses(jBonuses, &problem.m_bonuses);
        }

        if (problem.m_hole.size() > 108)
        {
            throw JException("Hole has too many vertices (see hard-coded arrays in edgeInPolygon)");
        }

        {
            uint64_t vertexCount = problem.m_vertices.size();
            vector<uint8_t> vertexUsed(vertexCount);
            for (auto& edge : problem.m_edges)
            {
                if (edge.a < 0 || edge.a >= vertexCount ||
                    edge.b < 0 || edge.b >= vertexCount)
                {
                    throw JException("Vertex index out of range");
                }
                vertexUsed[edge.a] = 1;
                vertexUsed[edge.b] = 1;
            }
            for (uint64_t iVertex = 0; iVertex < vertexCount; iVertex++)
            {
                if (!vertexUsed[iVertex])
                {
                    throw JException("Vertex not used by any edges");
                }
            }
        }

        {
            uint64_t vertexCount = problem.m_vertices.size();
            vector<uint8_t> vertexReachable(vertexCount);
            vertexReachable[0] = 1;
            bool didSomething = true;
            while (didSomething)
            {
                didSomething = false;
                for (auto& edge : problem.m_edges)
                {
                    if (vertexReachable[edge.a] &&
                        !vertexReachable[edge.b])
                    {
                        vertexReachable[edge.b] = 1;
                        didSomething = true;
                    }
                    if (vertexReachable[edge.b] &&
                        !vertexReachable[edge.a])
                    {
                        vertexReachable[edge.a] = 1;
                        didSomething = true;
                    }
                }
            }
            for (uint64_t iVertex = 0; iVertex < vertexCount; iVertex++)
            {
                if (!vertexReachable[iVertex])
                {
                    throw JException("Figure not connected");
                }
            }
        }

        for (size_t iEdge = 0; iEdge < problem.m_edges.size(); iEdge++)
        {
            if (problem.m_edges[iEdge].a == problem.m_edges[iEdge].b)
            {
                throw JException("Edge connects vertex to itself");
            }
#if 0
            // Apparently, duplicate edges are allowed
            for (size_t jEdge = 0; jEdge < problem.m_edges.size(); jEdge++)
            {
                if (jEdge == iEdge) continue;
                if ((problem.m_edges[iEdge].a == problem.m_edges[jEdge].a &&
                     problem.m_edges[iEdge].b == problem.m_edges[jEdge].b) ||
                    (problem.m_edges[iEdge].a == problem.m_edges[jEdge].b &&
                     problem.m_edges[iEdge].b == problem.m_edges[jEdge].a))
                {
                    throw JException("Duplicate edge");
                }
            }
#endif
        }

        for (uint64_t iCorner = 0; iCorner < problem.m_hole.size(); iCorner++)
        {
            for (uint64_t jCorner = 0; jCorner < iCorner; jCorner++)
            {
                if (problem.m_hole[iCorner] ==
                    problem.m_hole[jCorner])
                {
                    throw JException("Duplicate corner");
                }
            }
        }
    }

    void parseSolution(const J::Value& jSolution,
                       Solution* pSolution)
    {
        auto& solution = *pSolution;

        JCheckObject(jSolution);

        {
            auto& jVertices = JObjectGetArray(jSolution, "vertices");
            parsePoints(jVertices, &solution.m_vertices);
        }
    }

    template<typename Writer>
    void formatPoint(const Point& point,
                     Writer& w)
    {
        w.StartArray();
        w.Int64(point.x);
        w.Int64(point.y);
        w.EndArray();
    }

    template<typename Writer>
    void formatPoints(const vector<Point>& points,
                      Writer& w)
    {
        w.StartArray();
        for (auto& point: points)
        {
            formatPoint(point, w);
        }
        w.EndArray();
    }

    template<typename Writer>
    void formatEdge(const Edge& edge,
                    Writer& w)
    {
        w.StartArray();
        w.Int64(edge.a);
        w.Int64(edge.b);
        w.EndArray();
    }

    template<typename Writer>
    void formatEdges(const vector<Edge>& edges,
                     Writer& w)
    {
        w.StartArray();
        for (auto& edge: edges)
        {
            formatEdge(edge, w);
        }
        w.EndArray();
    }

    template<typename Writer>
    void formatProblem(const Problem& problem,
                       Writer& w)
    {
        w.StartObject();

        {
            w.Key("hole");
            formatPoints(problem.m_hole, w);
        }

        w.Key("epsilon");
        w.Uint64(problem.m_epsilon);

        w.Key("figure");
        w.StartObject();

        {
            w.Key("edges");
            formatEdges(problem.m_edges, w);
        }

        {
            w.Key("vertices");
            formatPoints(problem.m_vertices, w);
        }

        w.EndObject();

        w.EndObject();
    }

    template<typename Writer>
    void formatSolution(const Solution& solution,
                        Writer& w)
    {
        w.StartObject();

        {
            w.Key("vertices");
            formatPoints(solution.m_vertices, w);
        }

        w.EndObject();
    }
}

bool readProblem(uint64_t id,
                 string* pProblemData)
{
    string fileName = strprintf("../problems/%" PRIu64 "", id);
    return readFile(fileName, pProblemData);
}

bool writeProblem(uint64_t id,
                  const string& problemData)
{
    string fileName = strprintf("../problems/%" PRIu64 "", id);
    return writeFile(fileName, problemData);
}

bool parseProblem(const string& problemData,
                  Problem* pProblem,
                  string* pMsg)
{
    try
    {
        J::Document jProblem;
        jProblem.Parse(problemData);
        JCheckParseObject(jProblem);

        parseProblem(jProblem, pProblem);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseProblem(uint64_t id,
                  Problem* pProblem,
                  string* pMsg)
{
    string problemData;
    if (!readProblem(id, &problemData))
    {
        if (pMsg) *pMsg = strprintf("Error reading problem %" PRIu64 "", id);
        return false;
    }

    return parseProblem(problemData, pProblem, pMsg);
}

void formatProblem(const Problem& problem,
                   string* pProblemData)
{
    J::StringBuffer buffer;
    J::Writer<J::StringBuffer> w(buffer);

    formatProblem(problem, w);

    *pProblemData = buffer.GetString();
}

bool readSolution(uint64_t id,
                  string* pSolutionData)
{
    string fileName = strprintf("../solutions/%" PRIu64 "", id);
    return readFile(fileName, pSolutionData);
}

bool writeSolution(uint64_t id,
                   const string& solutionData)
{
    string fileName = strprintf("../solutions/%" PRIu64 "", id);
    return writeFile(fileName, solutionData);
}

bool parseSolution(const string& solutionData,
                   Solution* pSolution,
                   string* pMsg)
{
    try
    {
        J::Document jSolution;
        jSolution.Parse(solutionData);
        JCheckParseObject(jSolution);

        parseSolution(jSolution, pSolution);
    }
    catch (const JException& e)
    {
        if (pMsg) *pMsg = e.m_msg;
        return false;
    }

    return true;
}

bool parseSolution(uint64_t id,
                   Solution* pSolution,
                   string* pMsg)
{
    string solutionData;
    if (!readSolution(id, &solutionData))
    {
        if (pMsg) *pMsg = strprintf("Error reading solution %" PRIu64 "", id);
        return false;
    }

    return parseSolution(solutionData, pSolution, pMsg);
}

void formatSolution(const Solution& solution,
                    string* pSolutionData)
{
    J::StringBuffer buffer;
    J::Writer<J::StringBuffer> w(buffer);

    formatSolution(solution, w);

    *pSolutionData = buffer.GetString();
}
