#include "Common.hpp"
#include "ParseUtils.hpp"
#include "Problem.hpp"
#include "IsValid.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: analyze [<options>] <id>\n");
    printf("  <id>  Problem to analyze\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    uint64_t id = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string problemData;
    if (!readProblem(id, &problemData))
    {
        printf("Error reading problem\n");
        return 1;
    }

    string msg;
    Problem problem;
    if (!parseProblem(problemData, &problem, &msg))
    {
        printf("Error parsing problem %" PRIu64 ": %s\n", id, msg.c_str());
        return 1;
    }

    size_t numCorners = problem.m_hole.size();
    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
    {
        size_t jCorner = (iCorner + 1) % numCorners;
        printf("%3" PRIuZ " - %3" PRIuZ ":", iCorner, jCorner);

        uint64_t dNew = calcD(problem.m_hole[iCorner],
                              problem.m_hole[jCorner]);

        size_t numEdges = problem.m_edges.size();
        for (size_t iEdge = 0; iEdge < numEdges; iEdge++)
        {
            auto& edge = problem.m_edges[iEdge];
            uint64_t dOld = calcD(problem.m_vertices[edge.a],
                                  problem.m_vertices[edge.b]);
            if (isValidD(dOld, dNew, problem.m_epsilon))
            {
                printf(" [%" PRIuZ " %" PRIuZ "]", edge.a, edge.b);
            }
        }
        printf("\n");
    }

    size_t numVertices = problem.m_vertices.size();
    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
    {
        printf("%3" PRIuZ ":", iCorner);
        vector<uint8_t> prevs(numVertices);
        vector<uint8_t> nexts(numVertices);
        size_t prevCorner = (iCorner + numCorners - 1) % numCorners;
        size_t nextCorner = (iCorner + 1) % numCorners;

        uint64_t dNewPrev = calcD(problem.m_hole[iCorner],
                                  problem.m_hole[prevCorner]);
        uint64_t dNewNext = calcD(problem.m_hole[iCorner],
                                  problem.m_hole[nextCorner]);

        size_t numEdges = problem.m_edges.size();
        for (size_t iEdge = 0; iEdge < numEdges; iEdge++)
        {
            auto& edge = problem.m_edges[iEdge];
            uint64_t dOld = calcD(problem.m_vertices[edge.a],
                                  problem.m_vertices[edge.b]);
            if (isValidD(dOld, dNewPrev, problem.m_epsilon))
            {
                prevs[edge.a] = 1;
                prevs[edge.b] = 1;
            }
            if (isValidD(dOld, dNewNext, problem.m_epsilon))
            {
                nexts[edge.a] = 1;
                nexts[edge.b] = 1;
            }
        }

        for (size_t iVertex = 0; iVertex < numVertices; iVertex++)
        {
            if (prevs[iVertex] && nexts[iVertex])
            {
                printf(" %" PRIuZ "", iVertex);
            }
        }
        printf("\n");
    }

    size_t iCorner[4];
    for (iCorner[0] = 0; iCorner[0] < numCorners; iCorner[0]++)
    {
        for (size_t i = 1; i < 4; i++)
        {
            iCorner[i] = (iCorner[0] + i) % numCorners;
        }
        printf("%3" PRIuZ " %3" PRIuZ ":", iCorner[1], iCorner[2]);

        uint64_t dNew[3];
        for (size_t i = 0; i < 3; i++)
        {
            dNew[i] = calcD(problem.m_hole[iCorner[i]],
                            problem.m_hole[iCorner[i + 1]]);
        }

        size_t numEdges = problem.m_edges.size();
        for (size_t iEdge = 0; iEdge < numEdges; iEdge++)
        {
            auto& edgeI = problem.m_edges[iEdge];
            uint64_t dOldI = calcD(problem.m_vertices[edgeI.a],
                                   problem.m_vertices[edgeI.b]);
            if (!isValidD(dOldI, dNew[1], problem.m_epsilon)) continue;
            for (size_t jEdge = 0; jEdge < numEdges; jEdge++)
            {
                if (jEdge == iEdge) continue;
                auto& edgeJ = problem.m_edges[jEdge];
                uint64_t dOldJ = calcD(problem.m_vertices[edgeJ.a],
                                       problem.m_vertices[edgeJ.b]);
                if (!isValidD(dOldJ, dNew[0], problem.m_epsilon)) continue;
                for (size_t kEdge = 0; kEdge < numEdges; kEdge++)
                {
                    if (kEdge == iEdge || kEdge == jEdge) continue;
                    auto& edgeK = problem.m_edges[kEdge];
                    uint64_t dOldK = calcD(problem.m_vertices[edgeK.a],
                                           problem.m_vertices[edgeK.b]);
                    if (!isValidD(dOldK, dNew[2], problem.m_epsilon)) continue;
                    size_t vertices[3];
                    if (edgeJ.a == edgeI.a)
                    {
                        vertices[0] = edgeJ.b;
                        vertices[1] = edgeI.a;
                        vertices[2] = edgeI.b;
                    }
                    else if (edgeJ.b == edgeI.a)
                    {
                        vertices[0] = edgeJ.a;
                        vertices[1] = edgeI.a;
                        vertices[2] = edgeI.b;
                    }
                    else if (edgeJ.a == edgeI.b)
                    {
                        vertices[0] = edgeJ.b;
                        vertices[1] = edgeI.b;
                        vertices[2] = edgeI.a;
                    }
                    else if (edgeJ.b == edgeI.b)
                    {
                        vertices[0] = edgeJ.a;
                        vertices[1] = edgeI.b;
                        vertices[2] = edgeI.a;
                    }
                    else
                    {
                        continue;
                    }
                    if (edgeK.a == vertices[2])
                    {
                        vertices[3] = edgeK.b;
                    }
                    else if (edgeK.b == vertices[2])
                    {
                        vertices[3] = edgeK.a;
                    }
                    else
                    {
                        continue;
                    }
                    if (vertices[0] == vertices[2] ||
                        vertices[0] == vertices[3] ||
                        vertices[2] == vertices[3])
                    {
                        continue;
                    }
                    printf(" [%3" PRIuZ ", %3" PRIuZ ", %3" PRIuZ ", %3" PRIuZ "]", vertices[0], vertices[1], vertices[2], vertices[3]);
                }
            }
        }

        printf("\n");
    }

    return 0;
}
