#include "Common.hpp"
#include "Brute.hpp"
#include "Brute2.hpp"
#include "FastPoints.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "Best.hpp"
#include "DebugGraphics.hpp"
#include "Xoshiro.hpp"
#include <random>

using std::string;
using std::vector;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: brute [<options>] <id>\n");
    printf("  <id>  Problem to work on\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
    printf("  -g    Use debug graphics\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;

    bool help = false;
    bool saveBestChanges = true;
    bool useDebugGraphics = false;
    uint64_t id = 0;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveBestChanges = false;
        }
        else if (strArg == "-g")
        {
            useDebugGraphics = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;
    Problem problem;
    if (!parseProblem(id, &problem, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Best best;
    best.load(id);

    for (uint64_t pass = 2; pass < 3; pass++)
    {
        vector<Point> points;
        switch (pass)
        {
        case 0:
            printf("-------- Searching vertices: --------\n");
            points = problem.m_hole;
            break;
        case 1:
            printf("-------- Searching edges: --------\n");
            points = getPointsOnPolygon(problem.m_hole);
            break;
        case 2:
            printf("-------- Searching interiors: --------\n");
            points = getPointsInPolygon(problem.m_hole);
            break;
        default: ;
        }

        printf("Number of points: %" PRIuZ "\n", points.size());

        FastPoints fastPoints;
        fastPoints.init(problem, points);

        DebugGraphics debugGraphics(problem, fastPoints, points);
        string text;
        if (useDebugGraphics)
        {
            text = strprintf("Problem %" PRIu64 "", id);
            debugGraphics.create();
            debugGraphics.update(Solution(), text);
        }

#if 0
        {
            Gen orderGen(12348);
            std::shuffle(points.begin(), points.end(), orderGen);
        }
#endif

        uint64_t bestScore = (uint64_t)-1;
        uint64_t bestCount = 0;

        auto onSolution = [&](const Solution& solution,
                              uint64_t score) -> bool
        {
            bool better = (score < bestScore);
            bool same = (score == bestScore);
            bool newBest = better && best.update(solution, saveBestChanges);

            if (better)
            {
                bestScore = score;
                bestCount = 1;
            }
            else if (same)
            {
                bestCount++;
            }

            if (better || same)
            {
                printf("%7" PRIu64 "  %7" PRIu64 "%s\n",
                       bestCount,
                       score,
                       newBest ? "  BEST" :
                       better ? "  best" :
                       same ? "  same" : "");
            }

            return false;
        };

        auto onDebug = [&](const Solution& solution) -> bool
        {
            debugGraphics.update(solution, text);
            return debugGraphics.destroyed();
        };

        Brute2 brute(problem, fastPoints, points, onSolution,
                     useDebugGraphics ? onDebug : OnDebug(nullptr));
        brute.solve();

        if (useDebugGraphics)
        {
            debugGraphics.destroy();
        }
    }

    return 0;
}
