#include "Common.hpp"
#include "Brute.hpp"
#include "FastPoints.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "Best.hpp"
#include "DebugGraphics.hpp"
#include "Xoshiro.hpp"
#include <random>

using std::string;
using std::vector;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: improve [<options>] <id> [<seed1>] [<seed2>}\n");
    printf("  <id>  Problem to work on\n");
    printf("  <seed1>, <seed2>\n");
    printf("        Random number generator seeds (default: 1, 1)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
    printf("  -g    Use debug graphics\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotSeed1 = false;
    bool gotSeed2 = false;

    bool help = false;
    bool saveBestChanges = true;
    bool useDebugGraphics = false;
    uint64_t id = 0;
    uint32_t seed1 = 1;
    uint32_t seed2 = 1;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveBestChanges = false;
        }
        else if (strArg == "-g")
        {
            useDebugGraphics = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotSeed1 || !gotSeed2)
        {
            uint32_t seed = 0;
            if (!parseU32(strArg, &seed))
            {
                usage();
                return 1;
            }
            if (!gotSeed1)
            {
                seed1 = seed;
                gotSeed1 = true;
            }
            else
            {
                seed2 = seed;
                gotSeed2 = true;
            }
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;
    Problem problem;
    if (!parseProblem(id, &problem, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Best best;
    best.load(id);

    vector<Point> origPoints = getPointsInPolygon(problem.m_hole);
    vector<Point> points;

    printf("Number of points: %" PRIuZ "\n", origPoints.size());

    FastPoints fastPoints;
    fastPoints.init(problem, origPoints);

    DebugGraphics debugGraphics(problem, fastPoints, origPoints);
    string text;
    if (useDebugGraphics)
    {
        text = strprintf("Problem %" PRIu64 "", id);
        debugGraphics.create();
        debugGraphics.update(Solution(), text);
    }

    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    std::uniform_int_distribution<uint32_t> sameDist(0, 3);

    Solution solution;

    while (true)
    {
        std::seed_seq seq{seed1, seed2};
        Gen gen(seq);

        points = origPoints;
        std::shuffle(points.begin(), points.end(), gen);

        bool found = false;
        uint64_t score = 0;

        auto onSolution = [&](const Solution& solution_,
                              uint64_t score_) -> bool
        {
            solution = solution_;
            score = score_;
            found = true;
            return true;
        };

        Brute brute(problem, fastPoints, origPoints, onSolution, nullptr);
        brute.solve();

        if (found)
        {
            uint64_t didNothingCount = 0;
            while (didNothingCount < 4)
            {
                bool didSomething = false;
                for (size_t iPoint = 0; iPoint < points.size(); iPoint++)
                {
                    for (size_t iVertex = 0; iVertex < problem.m_vertices.size(); iVertex++)
                    {
                        if (solution.m_vertices[iVertex] == points[iPoint])
                        {
                            continue;
                        }
                        Point oldPoint = solution.m_vertices[iVertex];
                        solution.m_vertices[iVertex] = points[iPoint];
                        bool keep = false;
                        if (isValid(problem, solution))
                        {
                            uint64_t curScore = calcScore(problem, solution);
                            if (curScore < score)
                            {
                                keep = true;
                                score = curScore;
                                //printf("%7" PRIu64 "\n", score);
                                didSomething = true;
                            }
                            else if (curScore == score)
                            {
                                if (sameDist(gen) != 0)
                                {
                                    keep = true;
                                    score = curScore;
                                }
                            }
                        }
#if 1
                        if (keep && useDebugGraphics)
                        {
                            debugGraphics.update(solution, text);
                        }
#endif
                        if (!keep)
                        {
                            solution.m_vertices[iVertex] = oldPoint;
                        }
                    }
                }
                if (didSomething)
                {
                    didNothingCount = 0;
                }
                else
                {
                    didNothingCount++;
                }
            }

            bool better = (score < bestScore);
            bool same = (score == bestScore);
            bool newBest = better && best.update(solution, saveBestChanges);

            if (better)
            {
                bestScore = score;
                bestCount = 1;
#if 0
                if (useDebugGraphics)
                {
                    debugGraphics.update(problem, solution, text);
                }
#endif
            }
            else if (same)
            {
                bestCount++;
            }

            if (better || same)
            {
                printf("%10" PRIu32 "  %10" PRIu32 "  %7" PRIu64 "  %7" PRIu64 "%s\n",
                       seed1,
                       seed2,
                       bestCount,
                       score,
                       newBest ? "  BEST" :
                       better ? "  best" :
                       same ? "  same" : "");
            }
        }

        if (useDebugGraphics)
        {
            if (debugGraphics.destroyed())
            {
                break;
            }
        }

        seed2++;
        if (seed2 == 0)
        {
            seed2 = 1;
            seed1++;
            if (seed1 == 0)
            {
                break;
            }
        }
    }

    if (useDebugGraphics)
    {
        debugGraphics.destroy();
    }

    return 0;
}
