#include "Common.hpp"
#include "Brute.hpp"
#include "Brute2.hpp"
#include "FastPoints.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "Best.hpp"
#include "DebugGraphics.hpp"
#include "TimeUtils.hpp"
#include "Xoshiro.hpp"
#include <random>

using std::string;
using std::vector;

typedef Xoshiro256StarStar Gen;

void usage()
{
    printf("Usage: solve [<options>] <id> [<seed1>] [<seed2>}\n");
    printf("  <id>  Problem to work on\n");
    printf("  <seed1>, <seed2>\n");
    printf("        Random number generator seeds (default: 1, 1)\n");
    printf("Options:\n");
    printf("  -h    Print usage information and exit\n");
    printf("  -d    Don't save changes to best solutions\n");
    printf("  -g    Use debug graphics\n");
}

int main(int argc, char *argv[])
{
    bool gotId = false;
    bool gotSeed1 = false;
    bool gotSeed2 = false;

    bool help = false;
    bool saveBestChanges = true;
    bool useDebugGraphics = false;
    uint64_t id = 0;
    uint32_t seed1 = 1;
    uint32_t seed2 = 1;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg++];

        if (strArg == "-h" || strArg == "--help")
        {
            help = true;
        }
        else if (strArg == "-d")
        {
            saveBestChanges = false;
        }
        else if (strArg == "-g")
        {
            useDebugGraphics = true;
        }
        else if (!gotId)
        {
            if (!parseU64(strArg, &id) || id < 1 || id > 1000000)
            {
                usage();
                return 1;
            }
            gotId = true;
        }
        else if (!gotSeed1 || !gotSeed2)
        {
            uint32_t seed = 0;
            if (!parseU32(strArg, &seed))
            {
                usage();
                return 1;
            }
            if (!gotSeed1)
            {
                seed1 = seed;
                gotSeed1 = true;
            }
            else
            {
                seed2 = seed;
                gotSeed2 = true;
            }
        }
        else
        {
            usage();
            return 1;
        }
    }

    if (help)
    {
        usage();
        return 0;
    }

    if (!gotId)
    {
        usage();
        return 1;
    }

    string msg;
    Problem problem;
    if (!parseProblem(id, &problem, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Best best;
    best.load(id);

    vector<Point> points = getPointsInPolygon(problem.m_hole);

    printf("Number of points: %" PRIuZ "\n", points.size());

    FastPoints fastPoints;
    fastPoints.init(problem, points);

    DebugGraphics debugGraphics(problem, fastPoints, points);
    string text;
    if (useDebugGraphics)
    {
        text = strprintf("Problem %" PRIu64 "", id);
        debugGraphics.create();
        debugGraphics.update(Solution(), text);
    }

    uint64_t bestScore = (uint64_t)-1;
    uint64_t bestCount = 0;

    std::uniform_int_distribution<size_t> vertexDist(0, problem.m_vertices.size() - 1);
    std::uniform_int_distribution<size_t> holeDist(0, problem.m_hole.size() - 1);
    std::uniform_int_distribution<size_t> pointsDist(0, points.size() - 1);
    std::uniform_int_distribution<uint32_t> useHoleDist(0, 1);
    std::uniform_int_distribution<uint32_t> removeDist(0, 1);
    std::uniform_int_distribution<uint32_t> sameDist(0, 3);

    bool found = false;
    Solution curSolution;
    uint64_t curScore = 0;

    auto onSolution = [&](const Solution& solution_,
                          uint64_t score_) -> bool
    {
        found = true;
        curSolution = solution_;
        curScore = score_;
        return true;
    };

#if 0
    auto onDebug = [&](const Solution& solution) -> bool
    {
        if (useDebugGraphics)
        {
            debugGraphics.update(solution, text);
            //sleepMS(1000);
        }
        return false;
    };
#endif

    Brute brute(problem, fastPoints, points, onSolution, nullptr);

    uint64_t initMaxNodes = 0x1000;
    uint64_t maxNodes = 0x400;

    while (true)
    {
        std::seed_seq seq{seed1, seed2};
        Gen gen(seq);

        size_t firstVertex = vertexDist(gen);

        vector<Point> targets(problem.m_vertices.size());
#if 0
        for (size_t iVertex = 0; iVertex < problem.m_vertices.size(); iVertex++)
        {
            targets[iVertex] = points[pointsDist(gen)];
        }
#endif
#if 1
        for (size_t iVertex = 0; iVertex < problem.m_vertices.size(); iVertex++)
        {
            targets[iVertex] = Point(-999, -999);
        }
        if (useHoleDist(gen) == 0)
        {
            targets[firstVertex] = problem.m_hole[holeDist(gen)];
        }
        else
        {
            targets[firstVertex] = points[pointsDist(gen)];
        }
#endif
#if 0
        targets = problem.m_vertices;
#endif

        found = false;
        brute.solve(initMaxNodes, firstVertex, &targets);

        if (found)
        {
            Solution solution = curSolution;
            uint64_t score = curScore;

            if (useDebugGraphics)
            {
                debugGraphics.update(solution, text);
            }

            uint64_t didNothingCount = 0;
            uint64_t didNothingMax = problem.m_vertices.size() * 100;
            while (didNothingCount < didNothingMax)
            {
                bool didSomething = false;
                size_t iVertex = vertexDist(gen);
                Point oldTarget = targets[iVertex];
                if (oldTarget.x == -999 || removeDist(gen) == 0)
                {
                    if (useHoleDist(gen) == 0)
                    {
                        targets[iVertex] = problem.m_hole[holeDist(gen)];
                    }
                    else
                    {
                        targets[iVertex] = points[pointsDist(gen)];
                    }
                }
                else
                {
                    targets[iVertex] = Point(-999, -999);
                }

                found = false;
                brute.solve(maxNodes, firstVertex, &targets);

#if 0
                if (found)
                {
                    printf("curScore = %" PRIu64 "\n", curScore);
                }
                else
                {
                    printf("curScore = NOT FOUND\n");
                }
#endif

                bool keep = false;
                if (found)
                {
                    if (curScore < score)
                    {
                        keep = true;
                        didSomething = true;
#if 1
                        {
                            printf("\x1b[KcurScore = %" PRIu64 "\r", curScore);
                            fflush(stdout);
                        }
#endif
                    }
                    else if (curScore == score)
                    {
#if 1
                        if (sameDist(gen) != 0)
                        {
                            keep = true;
                        }
#endif
                    }
                }

                if (keep)
                {
                    solution = curSolution;
                    score = curScore;
                }
                else
                {
                    targets[iVertex] = oldTarget;
                }

                //if (keep)
                if (found)
                {
#if 1
                    if (useDebugGraphics)
                    {
                        debugGraphics.update(curSolution, text);
                    }
#endif
                }

                if (useDebugGraphics && debugGraphics.destroyed())
                {
                    break;
                }

                if (didSomething)
                {
                    didNothingCount = 0;
                }
                else
                {
                    didNothingCount++;
                }
            }

            bool better = (score < bestScore);
            bool same = (score == bestScore);
            bool newBest = better && best.update(solution, saveBestChanges);

            if (better)
            {
                bestScore = score;
                bestCount = 1;
#if 0
                if (useDebugGraphics)
                {
                    debugGraphics.update(solution, text);
                }
#endif
            }
            else if (same)
            {
                bestCount++;
            }

            if (better || same || true)
            {
                printf("%10" PRIu32 "  %10" PRIu32 "  %7" PRIu64 "  %7" PRIu64 "%s\n",
                       seed1,
                       seed2,
                       bestCount,
                       score,
                       newBest ? "  BEST" :
                       better ? "  best" :
                       same ? "  same" : "");
            }
        }
        else
        {
            if (true)
            {
                printf("%10" PRIu32 "  %10" PRIu32 "  NOT FOUND\n",
                       seed1,
                       seed2);
            }
        }

        if (useDebugGraphics && debugGraphics.destroyed())
        {
            break;
        }

        seed2++;
        if (seed2 == 0)
        {
            seed2 = 1;
            seed1++;
            if (seed1 == 0)
            {
                break;
            }
        }
    }

    if (useDebugGraphics)
    {
        debugGraphics.destroy();
    }

    return 0;
}
